/*********************************************************************

 $Id: RadiusStack.cxx,v 1.1 2000/06/09 03:10:18 fhong Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

////////////////////////////////////////////////////////////////////////
//
//  Parts of this code were copied from the free distribution available at:
//
//  Livingston Enterprises, Inc.
//  6920 Koll Center Parkway
//  Pleasanton, CA   94566
// 
//  Copyright 1992 Livingston Enterprises, Inc.
// 
//  Permission to use, copy, modify, and distribute this software for any
//  purpose and without fee is hereby granted, provided that this
//  copyright and permission notice appear on all copies and supporting
//  documentation, the name of Livingston Enterprises, Inc. not be used
//  in advertising or publicity pertaining to distribution of the
//  program without specific prior permission, and notice be given
//  in supporting documentation that copying and distribution is by
//  permission of Livingston Enterprises, Inc.   
// 
//  Livingston Enterprises, Inc. makes no representations about
//  the suitability of this software for any purpose.  It is
//  provided "as is" without express or implied warranty.
//
////////////////////////////////////////////////////////////////////////

 *********************************************************************

 $Log: RadiusStack.cxx,v $
 Revision 1.1  2000/06/09 03:10:18  fhong
 Abstracted the Mind VSA stuff from the Radius Stack


 *********************************************************************/

#include <sys/time.h>
#include <netinet/in.h>

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <pwd.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>

#include "RadiusStack.hxx"
#include "VRadiusException.hxx"
#include "NetworkAddress.h"
#include "UdpStack.hxx"
#include "vmd5.h"
#include "cpLog.h"

RadiusStack::RadiusStack() :
    _retries(5),
    _serverName("unknown"),
    _serverAddr(0),
    _serverPort(0),
    _clientName("localhost"),
    _clientAddr(0),
    _clientPort(0),
    _requestId(0),
    _responseId(0),
    _secretKey("testing123"),
    _connected(false),
    _acctStatusType(ACCT_OFF),
    _sendBufferLen(0),
    _recvBufferLen(0),
    _recvBufferValid(false),
    _networkAddr(0),
    _udpConnection(0)
{
}

RadiusStack::~RadiusStack()
{
    if(_connected)
    {
	if(!accountingOff())
	    cpLog(LOG_ALERT,
		  "Accounting off failed, closing connection anyways");
    }

    if(_networkAddr)
    {
	delete _networkAddr;
	_networkAddr = 0;
    }

    if(_udpConnection)
    {
	delete _udpConnection;
	_udpConnection = 0;
    }
}

void
RadiusStack::setupConnection()
{
    // set session specific values

    struct servent      *svp;
    struct hostent      *hent;
    unsigned short       svcport; // network byte order
    char                 nasname[256];

    // find server port

    svp = getservbyname ("radacct", "udp");
    if (svp ==  0)
	svcport = htons((unsigned short)PW_ACCT_UDP_PORT);
    else
	svcport = svp->s_port; // already in network byte order

    _serverPort = ntohs(svcport);

    if(gethostname(nasname, sizeof(nasname)) == 0)
	_clientName = nasname;
    else
    {
	string msg("Can't initialize radius connection, gethostname() failed");
	if(errno == EINVAL)
	    msg += " due to invalid name length";
	throw VRadiusException(msg,__FILE__,__LINE__);
    }

    if((hent = gethostbyname(nasname)) == 0)
    {
	string msg("Couldn't find IP for ");
	msg += nasname;
	throw VRadiusException(msg,__FILE__,__LINE__);
    }
    else
	_clientAddr = *((unsigned int*) hent->h_addr_list[0]); // network byte order

    // Get the IP address of the radius accounting server
    if((hent = gethostbyname(_serverName.c_str())) == 0)
    {
	string msg("Couldn't find IP for ");
	msg += _serverName;
	throw VRadiusException(msg,__FILE__,__LINE__);
    }
    else
	_serverAddr = *((unsigned int*) hent->h_addr_list[0]); // network byte order

    char buf[256];
    sprintf(buf,"Server Host:%s Server Port:%d",
	    _serverName.c_str(),_serverPort);
    cpLog(LOG_DEBUG,buf);
}

void
RadiusStack::initialize(const string &server,
			const string &secretKey,
			const int retries) throw(VRadiusException&)
{
    memset(&_sendBuffer,0,sizeof(_sendBuffer));
    memset(&_recvBuffer,0,sizeof(_recvBuffer));

    _serverName = server;
    _secretKey = secretKey;
    _retries = retries;

    setupConnection();

    try
    {
	if(_networkAddr !=0)
	    delete _networkAddr;

	if(_udpConnection !=0)
	    delete _udpConnection;

	_networkAddr = new NetworkAddress(_serverName,_serverPort);
	_udpConnection = new UdpStack(_networkAddr,_serverPort,_serverPort);
	
	// this may not be necessary, depends on the radius server
	_udpConnection->connectPorts();
    }
    catch(UdpStackException &e)
    {
	throw VRadiusException("UdpStackException caught",__FILE__,__LINE__);	
    }

    if(!accountingOn())
	throw VRadiusException("ACCOUNTING ON message failed, Radius server not responding",
			    __FILE__,__LINE__);
}

int RadiusStack::md5Calc(unsigned char *output,
			  unsigned char *input, unsigned inLen)
{
    MD5Context context;
    MD5Init(&context);
    MD5Update(&context, input, inLen);
    MD5Final(output, &context);

    return 0;
}

bool
RadiusStack::evaluateRecvBuffer()
{
    // evaluate attribute list in _receiveBuffer

    BufType::AcctHdr *auth = &(_recvBuffer.acctHdr);
    unsigned char    *ptr  = _recvBuffer.buffer;
    int               totalLen;
    int               count;
    unsigned char     attribType;
    unsigned char     attribLen;
    AttribType        enumAttribType;
    unsigned char     vsaLength;
    unsigned char     vsaType;

    if(!_recvBufferValid)
	return false;

    totalLen = ntohs(auth->length);
    count = sizeof(BufType::AcctHdr);

    char msgBuf[64];
    sprintf(msgBuf,"Received reply from radius server %s code=%d id=%d length=%d",
	    _serverName.c_str(),auth->code,auth->id,totalLen);
    cpLog(LOG_DEBUG,msgBuf);

    // fast forward to the start of the attributes
    ptr += count;

    while(count < totalLen)
    {
	attribType = (unsigned char)*ptr++;
	attribLen  = (unsigned char)*ptr++;
	count += attribLen;
	enumAttribType = (AttribType) attribType;

	// check if attribute value is empty
	if(attribLen < 2)
	    continue;

	char logMsg[128];
	switch(enumAttribType) {

	    case PW_ACCT_STATUS_TYPE:
		_acctStatusType = (AcctStatusType)ntohl((int)*ptr);
		sprintf(logMsg,"Attribute PW_ACCT_STATUS_TYPE:%d",
			_acctStatusType);
		cpLog(LOG_DEBUG,logMsg);
		ptr += attribLen - 2;
		break;

	    case PW_VENDOR_SPECIFIC:
		ptr += 4;
		vsaType = (unsigned char)*ptr++;
		vsaLength = (unsigned char)*ptr++;
		processVsa(vsaType,ptr,vsaLength);
		ptr += vsaLength - 2;
		break;

	    default:
		sprintf(logMsg,"Ignored attribute type:%d, length:%d",
			(int)enumAttribType,
			(int)attribLen);
		cpLog(LOG_DEBUG,logMsg);
		ptr += attribLen - 2;
		break;
	}
    }
    return true;
}

bool
RadiusStack::verifyAcctRecvBuffer() throw(VRadiusException&)
{
    BufType             bufCopy;
    BufType::AcctHdr   *auth = &(bufCopy.acctHdr);
    unsigned char      *ptr = bufCopy.buffer;
    int                 totalLen;
    unsigned char       reply_digest[ACCT_VECTOR_LEN];
    unsigned char       calc_digest[ACCT_VECTOR_LEN];

    // initialize validity of _recvBuffer to false
    _recvBufferValid = false;

    // make a copy of the radius message
    memcpy(&bufCopy,&_recvBuffer,sizeof(bufCopy));

    totalLen = ntohs(auth->length);
    _responseId = auth->id;

    if(totalLen != _recvBufferLen)
    {
	char buf[64];
	sprintf(buf,
		"Received invalid reply length from server (want %d got %d)",
		totalLen,_recvBufferLen);
	throw VRadiusException(buf,__FILE__,__LINE__);
    }

    // zero out the auth vector in the message in order to calculate the
    // correct md5 sum
    memcpy(reply_digest, auth->authVector, ACCT_VECTOR_LEN);
    memset(auth->authVector,0,ACCT_VECTOR_LEN);

    // Verify the reply digest
    memcpy(ptr + totalLen, _secretKey.c_str(), _secretKey.size());
    md5Calc(calc_digest, (unsigned char *)auth,
	    totalLen + _secretKey.size());

    if(memcmp(reply_digest, calc_digest, ACCT_VECTOR_LEN) != 0)
    {
	cpLog(LOG_DEBUG,"Warning: Received invalid reply digest from server");
    }
    else
    {
	// check that the packet type is PW_ACCOUNTING_RESPONSE and
	// the identifier matches the request identifier
	if((PacketType)auth->code == PW_ACCOUNTING_RESPONSE &&
	   _responseId == _requestId)
	    _recvBufferValid = true;
    }

    if(_recvBufferValid)
	return(evaluateRecvBuffer());
    else
	return _recvBufferValid;
}

bool
RadiusStack::handshake()
{
    int result = 0;
    int i = 0;
    for (; i < _retries; i++)
    {
	char buf[32];
	sprintf(buf,"Sending request. Attempt #%d",i+1);
	cpLog(LOG_DEBUG,buf);

	if(!sendRadius())
	    return false;

	if((result = recvRadius()) > 0) 
	{
	    cpLog(LOG_DEBUG,"Received a reply from Radius");
	    break;
	}
    }
    if (result > 0 && i < _retries)
    {
	_connected = true;
	return _connected;
    }
    _connected = false;
    return false;
}

bool
RadiusStack::sendRadius()
{
    try
    {
	_udpConnection->transmit((char*)&(_sendBuffer.buffer),
				 _sendBufferLen);
    }
    catch(UdpStackException &e)
    {
	cpLog(LOG_ALERT,"Error occured in sendRadius()");
	cpLog(LOG_ALERT,"UdpStackException caught");
	return false;
    }

    char buf[64];
    sprintf(buf,"Transmitted bytes:%d",_sendBufferLen);
    cpLog(LOG_DEBUG,buf);

    return true;
}

int 
RadiusStack::recvRadius()
{
    struct timeval   tv;
    fd_set           readfds;

    tv.tv_sec = 5;
    tv.tv_usec = 1;

    FD_ZERO(&readfds);
    FD_SET(_udpConnection->getSocketFD(), &readfds);

    if (select(_udpConnection->getSocketFD()+1, &readfds, 0, 0, &tv) == 0)
        return(0);

    try
    {
	_recvBufferLen = _udpConnection->receive((char*)&(_recvBuffer.buffer),
						 sizeof(_recvBuffer.buffer));
    }
    catch(UdpStackException &e)
    {
	cpLog(LOG_ALERT,"Error occured in recvRadius()");
	cpLog(LOG_ALERT,"UdpStackException caught");
	return 0;
    }
    
    return(_recvBufferLen);
}

int
RadiusStack::addHeader(const PacketType type)
{
    BufType::AcctHdr *auth = &(_sendBuffer.acctHdr);

    memset(&_sendBuffer,0,sizeof(_sendBuffer));

    auth->code = type;
    auth->id = getpid() % 256;

    _requestId = auth->id;
    _sendBufferLen = sizeof(BufType::AcctHdr);

    return _sendBufferLen;
}

/**
   Algorithm for Request/Response authenticator:
   MD5(code + id + length + authenticator + attributes + secretKey)
**/
void
RadiusStack::addMD5()
{
    BufType::AcctHdr *auth = &(_sendBuffer.acctHdr);
    unsigned char    *ptr  = _sendBuffer.buffer;

    ptr += _sendBufferLen;

    // Secret key, used to create digest, do not send secret key
    strcpy((char*)ptr,_secretKey.c_str());
    int bufSizeWithKey  = _sendBufferLen + _secretKey.size();

    // Set the length field
    auth->length = htons(_sendBufferLen);

    // set authenticator to zeros before calculating hash
    memset(auth->authVector, 0, ACCT_VECTOR_LEN);

    // Calculate the MD5 Digest of the entire message
    unsigned char digest[ACCT_VECTOR_LEN];
    md5Calc(digest,_sendBuffer.buffer, bufSizeWithKey);

    memcpy(auth->authVector,digest,sizeof(digest));
}

int
RadiusStack::addAttribute(const AttribType type, const unsigned int value)
{
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeof(unsigned int) + 2;

    ptr += _sendBufferLen;
    *ptr++ = (unsigned char)type;
    *ptr++ = length;
    memcpy(ptr, &value, sizeof(unsigned int));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addAttribute(const AttribType type, const unsigned short value)
{
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeof(unsigned short) + 2;

    ptr += _sendBufferLen;
    *ptr++ = (unsigned char)type;
    *ptr++ = length;
    memcpy(ptr, &value, sizeof(unsigned short));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addAttribute(const AttribType type, const unsigned char value)
{
    unsigned char *ptr  = _sendBuffer.buffer;
    char           length = sizeof(unsigned char) + 2;

    ptr += _sendBufferLen;
    *ptr++ = (unsigned char)type;
    *ptr++ = length;
    memcpy(ptr, &value, sizeof(unsigned char));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addAttribute(const AttribType type,
			  const unsigned char *value,
			  const int sizeOfValue)
{
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeOfValue + 2;

    ptr += _sendBufferLen;
    *ptr++ = (unsigned char)type;
    *ptr++ = length;
    if(sizeOfValue > 0 && ptr)
	memcpy(ptr, value, sizeOfValue);
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addVsaAttribute(const unsigned char vsaType,
			     const unsigned long value,
			     const unsigned int vendorId)
{
    unsigned int  *intPtr;
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeof(unsigned long) + 8;

    ptr += _sendBufferLen;

    *ptr++ = (unsigned char)PW_VENDOR_SPECIFIC; // RAD Type
    *ptr++ = length;                            // RAD length
    intPtr = (unsigned int*)ptr;
    *intPtr = htonl(vendorId);                  // Vendor ID
    ptr += 4;
    *ptr++ = (unsigned char)vsaType;               // Vendor Type
    *ptr++ = sizeof(unsigned long) + 2;         // Vendor Length

    memcpy(ptr, &value, sizeof(unsigned long));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addVsaAttribute(const unsigned char vsaType,
			     const unsigned int value,
			     const unsigned int vendorId)
{
    unsigned int  *intPtr;
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeof(unsigned int) + 8;

    ptr += _sendBufferLen;

    *ptr++ = (unsigned char)PW_VENDOR_SPECIFIC; // RAD Type
    *ptr++ = length;                            // RAD length
    intPtr = (unsigned int*)ptr;
    *intPtr = htonl(vendorId);                  // Vendor ID
    ptr += 4;
    *ptr++ = (unsigned char)vsaType;               // Vendor Type
    *ptr++ = sizeof(unsigned int) + 2;          // Vendor Length

    memcpy(ptr, &value, sizeof(unsigned int));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addVsaAttribute(const unsigned char vsaType,
			     const unsigned short value,
			     const unsigned int vendorId)
{
    unsigned int  *intPtr;
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeof(unsigned short) + 8;

    ptr += _sendBufferLen;

    *ptr++ = (unsigned char)PW_VENDOR_SPECIFIC; // RAD Type
    *ptr++ = length;                            // RAD length
    intPtr = (unsigned int*)ptr;
    *intPtr = htonl(vendorId);                  // Vendor ID
    ptr += 4;
    *ptr++ = (unsigned char)vsaType;               // Vendor Type
    *ptr++ = sizeof(unsigned short) + 2;        // Vendor Length

    memcpy(ptr, &value, sizeof(unsigned short));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addVsaAttribute(const unsigned char vsaType,
			     const unsigned char value,
			     const unsigned int vendorId)
{
    unsigned int  *intPtr;
    unsigned char *ptr  = _sendBuffer.buffer;
    char           length = sizeof(unsigned char) + 8;

    ptr += _sendBufferLen;

    *ptr++ = (unsigned char)PW_VENDOR_SPECIFIC; // RAD Type
    *ptr++ = length;                            // RAD length
    intPtr = (unsigned int*)ptr;
    *intPtr = htonl(vendorId);                  // Vendor ID
    ptr += 4;
    *ptr++ = (unsigned char)vsaType;               // Vendor Type
    *ptr++ = sizeof(unsigned char) + 2;         // Vendor Length

    memcpy(ptr, &value, sizeof(unsigned char));
    _sendBufferLen += length;

    return length;
}

int
RadiusStack::addVsaAttribute(const unsigned char vsaType,
			     const unsigned char *value,
			     const int sizeOfValue,
			     const unsigned int vendorId)
{
    unsigned int  *intPtr;
    unsigned char *ptr  = _sendBuffer.buffer;
    unsigned char  length = sizeOfValue + 8;

    ptr += _sendBufferLen;

    *ptr++ = (unsigned char)PW_VENDOR_SPECIFIC; // RAD Type
    *ptr++ = length;                            // RAD length
    intPtr = (unsigned int*)ptr;
    *intPtr = htonl(vendorId);                  // Vendor ID
    ptr += 4;
    *ptr++ = (unsigned char)vsaType;               // Vendor Type
    *ptr++ = (unsigned char)sizeOfValue + 2;    // Vendor Length

    if(sizeOfValue > 0 && ptr)
	memcpy(ptr, value, sizeOfValue);
    _sendBufferLen += length;

    return length;
}

void
RadiusStack::createAcctOnMsg()
{
    addHeader(PW_ACCOUNTING_REQUEST);
    addAttribute(PW_ACCT_STATUS_TYPE,htonl(ACCT_ON));
    addAttribute(PW_NAS_IP_ADDRESS,_clientAddr);
    addMD5();
}

void
RadiusStack::createAcctOffMsg()
{
    addHeader(PW_ACCOUNTING_REQUEST);
    addAttribute(PW_ACCT_STATUS_TYPE,htonl(ACCT_OFF));
    addAttribute(PW_NAS_IP_ADDRESS,_clientAddr);
    addMD5();
}

void
RadiusStack::createAcctStartCallMsg(const string &from,
				    const string &to,
				    const string &callId)
{
    addHeader(PW_ACCOUNTING_REQUEST);

    addAttribute(PW_ACCT_STATUS_TYPE,htonl(ACCT_START));
    addAttribute(PW_ACCT_SESSION_ID,(unsigned char*)callId.c_str(),
		 callId.size());

    addAttribute(PW_CALLING_STATION_ID,
		 (unsigned char*)from.c_str(),
		 from.size());
    addAttribute(PW_CALLED_STATION_ID,
		 (unsigned char*)to.c_str(),
		 to.size());
    addAttribute(PW_USER_NAME,(unsigned char*)callId.c_str(),
		 callId.size());
    
    addAttribute(PW_NAS_IP_ADDRESS,_clientAddr);
    
    addMD5();
}

void
RadiusStack::createAcctStopCallMsg(const string &callId)
{
    addHeader(PW_ACCOUNTING_REQUEST);

    addAttribute(PW_ACCT_STATUS_TYPE,htonl(ACCT_STOP));
    addAttribute(PW_ACCT_SESSION_ID,(unsigned char*)callId.c_str(),
		 callId.size());
    addMD5();
}

void
RadiusStack::createAccessRqstMsg(const string &callId,
				 const string &passwd)
{
    addHeader(PW_AUTHENTICATION_REQUEST);

    // TODO Password still needs to be hidden using
    // TODO RSA message Digest Algorithm MD5
    //
    addAttribute(PW_PASSWORD,(unsigned char*)passwd.c_str(),
		 passwd.size());

    addAttribute(PW_USER_NAME,(unsigned char*)callId.c_str(),
		 callId.size());
    addAttribute(PW_NAS_IP_ADDRESS,_clientAddr);

    addMD5();
}

bool
RadiusStack::accountingOn()
{
    cpLog(LOG_INFO,"Requesting Accounting On");
    createAcctOnMsg();
    if(handshake())
	return verifyAcctRecvBuffer();
    return false;
}

bool
RadiusStack::accountingOff()
{
    cpLog(LOG_INFO,"Requesting Accounting Off");
    createAcctOffMsg();
    if(handshake())
	return verifyAcctRecvBuffer();
    return false;
}

bool
RadiusStack::accountingStartCall(const string &from,
				 const string &to,
				 const string &callId)
{
    if(!_connected)
    {
	if(!accountingOn())
	    return false;
    }
    cpLog(LOG_INFO,"Requesting Accounting Start Call");
    createAcctStartCallMsg(from,to,callId);
    if(handshake())
	return verifyAcctRecvBuffer();
    return false;
}

bool
RadiusStack::accountingStopCall(const string &callId)
{
    if(!_connected)
    {
	if(!accountingOn())
	    return false;
    }
    cpLog(LOG_INFO,"Requesting Accounting Stop Call");
    createAcctStopCallMsg(callId);
    if(handshake())
	return verifyAcctRecvBuffer();
    return false;
}

bool
RadiusStack::accessRequest(const string &callId,
			   const string &passwd)
{
    cpLog(LOG_INFO,"Requesting Authentication and Access");
    createAccessRqstMsg(callId,passwd);
    if(handshake())
	return verifyAuthRecvBuffer();
    return false;
}
