/*********************************************************************

 $Id: RadiusStack.hxx,v 1.1 2000/06/09 03:10:18 fhong Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

////////////////////////////////////////////////////////////////////////
//
//  Parts of this code were copied from the free distribution available at:
//
//  Livingston Enterprises, Inc.
//  6920 Koll Center Parkway
//  Pleasanton, CA   94566
// 
//  Copyright 1992 Livingston Enterprises, Inc.
// 
//  Permission to use, copy, modify, and distribute this software for any
//  purpose and without fee is hereby granted, provided that this
//  copyright and permission notice appear on all copies and supporting
//  documentation, the name of Livingston Enterprises, Inc. not be used
//  in advertising or publicity pertaining to distribution of the
//  program without specific prior permission, and notice be given
//  in supporting documentation that copying and distribution is by
//  permission of Livingston Enterprises, Inc.   
// 
//  Livingston Enterprises, Inc. makes no representations about
//  the suitability of this software for any purpose.  It is
//  provided "as is" without express or implied warranty.
//
////////////////////////////////////////////////////////////////////////

 *********************************************************************

 $Log: RadiusStack.hxx,v $

 *********************************************************************/
#ifndef  RadiusStack_hxx
#define RadiusStack_hxx

#include <sys/types.h>
#include <sys/socket.h>
#include <string>

/// Global constants

const int ACCT_VECTOR_LEN = 16;
const int PW_AUTH_UDP_PORT = 1812;
const int PW_ACCT_UDP_PORT = 1813;

class VRadiusException;
class NetworkAddress;
class UdpStack;

/// Data structure definitions

union BufType  /// Used for send and receive buffers
{
    /// attribute buffer
    unsigned char buffer[4096];
    
    /// header for attribute buffer (attributes must follow header)
    struct AcctHdr
    {
	unsigned char  code;
	unsigned char  id;
	unsigned short length;
	unsigned char  authVector[ACCT_VECTOR_LEN];
    } acctHdr;
};


/** 
    RadiusStack communicates with a radius server based on
    the radius protocol (See RFC 2139)

    This implementation is meant to work as a base class for
    a client application. The client should define it's own
    methods to deal with different VSA's (Vendor Specific
    Attributes).

    NOTE: most of the work has been done on the accounting
    functionality, there is little support for the
    authorization functionality.
**/
class RadiusStack
{
public:
    enum PacketType
    {
	PW_AUTHENTICATION_REQUEST = 1,
	PW_AUTHENTICATION_ACK     = 2,
	PW_AUTHENTICATION_REJECT  = 3,
	PW_ACCOUNTING_REQUEST     = 4,
	PW_ACCOUNTING_RESPONSE    = 5,
	PW_ACCOUNTING_STATUS      = 6,
	PW_PASSWORD_REQUEST       = 7,
	PW_PASSWORD_ACK           = 8,
	PW_PASSWORD_REJECT        = 9,
	PW_ACCOUNTING_MESSAGE     = 10,
	PW_ACCESS_CHALLENGE       = 11
    };

    enum AttribType
    {
	PW_USER_NAME            = 1,
	PW_PASSWORD             = 2,
	PW_CHAP_PASSWORD        = 3,
	PW_NAS_IP_ADDRESS       = 4,
	PW_NAS_PORT_ID          = 5,

	PW_VENDOR_SPECIFIC      = 26,
	PW_CALLED_STATION_ID	= 30,
	PW_CALLING_STATION_ID   = 31,
	
	PW_ACCT_STATUS_TYPE     = 40,
	PW_ACCT_DELAY_TIME      = 41,
	PW_ACCT_INPUT_OCTETS    = 42,
	PW_ACCT_OUTPUT_OCTETS   = 43,
	PW_ACCT_SESSION_ID      = 44,
	PW_ACCT_AUTHENTIC       = 45,
	PW_ACCT_SESSION_TIME    = 46,
	PW_ACCT_INPUT_PACKETS   = 47,
	PW_ACCT_OUTPUT_PACKETS  = 48,
	PW_ACCT_TERMINATE_CAUSE = 49
    };

    enum AcctStatusType
    {
	ACCT_START  = 1,
	ACCT_STOP   = 2,
	ACCT_ON     = 7,
	ACCT_OFF    = 8
    };

public:

    /// Constructor
    RadiusStack();
    
    /// Destructor
    virtual ~RadiusStack();

    /// Initialize connection with radius server
    void initialize(const string &server,
		    const string &secretKey,
		    const int     retries) throw(VRadiusException&);

    /// Set up the values for the connection
    void setupConnection();

    /// Sent when the client starts
    virtual bool accountingOn();

    /// Sent when the client terminates
    virtual bool accountingOff();

    /// Sent when subscriber should be authenticated
    virtual bool accessRequest(const string &callId,
			       const string &passwd);

    /// Sent when gateway wants to start a call
    virtual bool accountingStartCall(const string &from,
				     const string &to,
				     const string &callId);

    /// Sent when the call is terminated
    virtual bool accountingStopCall(const string &callId);


protected:
    /** Protected methods **/

    /** Process the VSA (vendor specific attributes)
	Called by evaluateRecvBuffer() to deal with
	messages from the radius server
    **/
    virtual void processVsa(const unsigned char vsaType,
			    unsigned char *ptr,
			    int bufLen) {};

    /// Fills the _sendBuffer with accounting ON message
    virtual void createAcctOnMsg();

    /// Fills the _sendBuffer with accounting OFF message
    virtual void createAcctOffMsg();

    /// Fills the _sendBuffer with accounting start call message
    virtual void createAcctStartCallMsg(const string &from,
					const string &to,
					const string &callId);
 
    /// Fills the _sendBuffer with accounting stop call message
    virtual void createAcctStopCallMsg(const string &callId);
    
    /// Fills the _sendBuffer with access request  message
    virtual void createAccessRqstMsg(const string &callId,
				     const string &passwd);

    /// Fill _sendBuffer header with data, overwrites existing data
    int addHeader(const PacketType type);

    /// Calculate MD5 hash and fill field in _sendBuffer
    void addMD5();

    /// Adds attribute to _sendBuffer, returns length of attribute
    int addAttribute(const AttribType type, const unsigned int value);

    /// Adds attribute to _sendBuffer, returns length of attribute
    int addAttribute(const AttribType type, const unsigned short value);

    /// Adds attribute to _sendBuffer, returns length of attribute
    int addAttribute(const AttribType type, const unsigned char value);

    /// Adds attribute to _sendBuffer, returns length of attribute
    int addAttribute(const AttribType type,
		     const unsigned char* value,
		     const int sizeOfValue);

    /// Adds vsaAttribute to _sendBuffer, returns length of vsaAttribute
    int addVsaAttribute(const unsigned char vsaType,
			const unsigned long value,
			const unsigned int vendorId);

    /// Adds vsaAttribute to _sendBuffer, returns length of vsaAttribute
    int addVsaAttribute(const unsigned char vsaType,
			const unsigned int value,
			const unsigned int vendorId);

    /// Adds vsaAttribute to _sendBuffer, returns length of vsaAttribute
    int addVsaAttribute(const unsigned char vsaType,
			const unsigned short value,
			const unsigned int vendorId);

    /// Adds vsaAttribute to _sendBuffer, returns length of vsaAttribute
    int addVsaAttribute(const unsigned char vsaType,
			const unsigned char value,
			const unsigned int vendorId);

    /// Adds vsaAttribute to _sendBuffer, returns length of vsaAttribute
    int addVsaAttribute(const unsigned char vsaType,
			const unsigned char* value,
			const int sizeOfValue,
			const unsigned int vendorId);

    /// sends message and waits for reply
    bool handshake();

    /// Send data from _sendBuffer. Blocks until all the data is sent
    bool sendRadius();

    /// Puts data in _recvBuffer
    int recvRadius();
    
    /// evaluate attribute list in receive _recvBuffer
    bool evaluateRecvBuffer();

    /// validate md5 sum of accounting message in _recvBuffer
    bool verifyAcctRecvBuffer() throw(VRadiusException&);

    /// validate md5 sum of authentication message in _recvBuffer
    bool verifyAuthRecvBuffer() throw(VRadiusException&) {return(true);}

    /// md5Calc is used to find the md5 sum
    int md5Calc(unsigned char *output, unsigned char *input, unsigned inLen);


protected:

    /** Data members **/

    int              _retries;

    string           _serverName;
    unsigned int     _serverAddr;     /// network byte order
    int              _serverPort;

    string           _clientName;
    unsigned int     _clientAddr;     /// network byte order
    int              _clientPort;

    unsigned char    _requestId;
    unsigned char    _responseId;
    string           _secretKey;
    
    bool             _connected;
    AcctStatusType   _acctStatusType;

    BufType          _sendBuffer;
    BufType          _recvBuffer;
    int              _sendBufferLen;
    int              _recvBufferLen;
    bool             _recvBufferValid;

    NetworkAddress  *_networkAddr;
    UdpStack        *_udpConnection;
};
#endif
