/*********************************************************************

 $Id: ControlEvent.hxx,v 1.1 2000/05/27 07:14:36 icahoon Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************

 $Log: ControlEvent.hxx,v $
 Revision 1.1  2000/05/27 07:14:36  icahoon
 Event to control the starting, stopping and shutting down of a thread.

 ********************************************************************/

#if !defined(CONTROL_EVENT_DOT_H)
#define CONTROL_EVENT_DOT_H

#include "Event.hxx"
#include "Sptr.hxx"
#include "Fifo.h"
#include "VMutex.h"

/// ControlEvent is used to manage the starting, stopping and shutting down of
/// thread that wait on Fifo<Event>'s. It is best to only use these within
/// an Sptr.
///
class ControlEvent : public Event
{
    public:

    	/// The type of control event.
	///
    	enum Type
	{
	    START = 0,
	    STOP,
	    SHUTDOWN
	};

    	/// Create a control event with a type and an optional reply
	/// fifo. If the reply fifo is set, the control event can be
	/// used as a request (rather than a command), and the response
	/// can be communicated via the returncode.
	///
    	ControlEvent(Type, Fifo< Sptr<Event> > * = 0);
    	ControlEvent(Type, Fifo< Sptr<Event> > &);

	virtual     ~ControlEvent();


    	virtual const char* const   	name() const;


    	Type	    	    	    	getType() const;
		

    	void	    	    	    	setReturnCode(int);
    	int 	    	    	    	getReturnCode() const;
		
    	void	    	    	    	setReplyFifo(
	    	    	    	    	    Fifo< Sptr<Event> > &
					);
    	Fifo< Sptr<Event> > * 	    	getReplyFifo() const;

    	/// Sends the event back to the reply fifo, if set.
	///
    	/// WARNING: This assumes that this is already housed in an Sptr.
	///  	     Don't use this method if it is on the stack, or it can be
	/// 	     manually deleted, since it will cause a double delete.
	/// 	
    	virtual void	    	    	returnReply();
			
    	/// To uniquely identify a request. Useful if this control event is
	/// a request. The fifo waiting for the response can the uniquely
	/// identify the response, in the case they sent out multiple control
	/// events simultaneously.
	///
    	int 	    	    	    	getControlEventId() const;


    private:
    	/// nextId is a globally shared resource, so it needs to be protected.
	///
    	static	VMutex	    	    	nextIdMutex;
    	static	int 	    	    	nextId;

	Type	    	    	    	type;
        int 	    	    	    	returnCode;
	Fifo< Sptr<Event> > 	    *	fifo;
    	int 	    	    	    	id;

    	/// Suppress copying.
	///
    	ControlEvent(const ControlEvent &);
	ControlEvent & operator=(const ControlEvent &);
};

#endif // !defined(CONTROL_EVENT_DOT_H)
