/*********************************************************************

 $Id: Fifo.h,v 1.16 2000/03/23 01:34:04 vincent Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************

 $Log: Fifo.h,v $
 Revision 1.16  2000/03/23 01:34:04  vincent
 Make nextEventID a static variable as there are multiple fifo objects that share the same queue and multiple events in the queue had the same eventId.

 Revision 1.15  2000/01/12 02:04:19  wenqing
 Modify the operator- of VTime to return a signed number instead of unsigned

 Revision 1.14  2000/01/11 00:56:00  bogawa
 copyright notice fixes for year 2000

 Revision 1.13  1999/12/30 20:40:24  eckelcu
 Modified Files:
 	Fifo.cxx Fifo.h VTime.cxx VTime.hxx
 Cleaned up and added comments to VTime class,
 and modified Fifo to use VTime class.

    
 ********************************************************************/

#ifndef FIFO_H
#define FIFO_H

#include <errno.h>
#include <list>
#include "VThread.h"
#include "VMutex.h"
#include "VCondition.h"
#include "VTime.hxx"
#include <unistd.h>
#include "debug.h"
#include "global.h"
    
typedef unsigned int FifoEventId;

/// Base class fof FIFO that implements the timed events functionality
class FifoBase
{
   protected:
      /// Message to hold the infromation about an event 
      struct FifoEvent
      {
	/// time this event should execute
	VTime time;
    
	/// id of this event 
	FifoEventId eventId;
	
	/// fifo to put the message on  
	FifoBase* fifo;
	
	/// pointer to heap message to put on FIFO 
	void* msg;
      };
       
 public:
      /// cancel a delayed message 
      void cancel( FifoEventId id );

 protected:
      /// constructor only for derived classes
      FifoBase();

      /// add the event to the event list 
      FifoEventId addEvent(FifoEvent& event, const int delayMs);

      /// called when a event is trigger to add the message to the Fifo
      virtual void add(const FifoEvent& msg) = 0;

      /// called to delete the msg stored in the FifoEvent
      virtual void deleteMsg(FifoEvent& msg) = 0;

      /// Mutex to protect access to data 
      VMutex *mutex;

      /// Conditional variable for waiting until the FIFO is not empty 
      VCondition *condition;

   private:
      /// value of next unused event ID 
      static FifoEventId nextEventID;

      /// list of all pending events in order of occurnace 
      static list<FifoEvent> events;
      
      /// control to make sure only one timer thread is started 
      static VMutex* onceControl;
      
      /// mutex to protect the events data structure 
      static VMutex* eventsMutex;
      
      /// main function for the timer thread 
      static void* timerThreadMain(void*);
      
      /// called when first event added to start the timer thread 
      static void startTimerThread();

      /// paramter that controls the resolution of the events 
      static const int maxErrorInMs;
};

    
/// compares two event based on the time they occur 
bool operator < ( FifoBase::FifoEvent& lhs, FifoBase::FifoEvent& rhs);


/// This template class implements a thread safe FIFO of messages
template<class Msg> class Fifo  : public FifoBase
{
   public:
      /// Create an empty FIFO
      Fifo();

      /// Delete the FIFO - must be empty when deleted 
      virtual ~Fifo();
      
      /// Add a new message to the FIFO 
      void add( const Msg& msg);

      /// add message to the FIFO delayMs miliseconds into the future 
      FifoEventId addDelayMs(const Msg& msg, const int delayMs);

      /// Get the next message from the FIFO - blocks until FIFO is not empty
      Msg getNext();

      /// Get the current size of the FIFO 
      int size() const;
      
   private:
      friend FifoEvent;
      
      /// add the message contained in the event to the Fifo
      virtual void add(const FifoEvent& event);
      
      /// delete the message contained in the event from the Fifo
      virtual void deleteMsg(FifoEvent& event);

      /// linked list structure holding all the message in the FIFO 
      list< Msg > data;

};


template<class Msg> 
Fifo<Msg>::Fifo( ) 
{
   // initialize all the members 
   mutex = new VMutex();
   assert(mutex);
   condition = new VCondition();
   assert(condition);
   assert( data.size() == 0 );
}


template<class Msg> 
Fifo<Msg>::~Fifo( )
{
   // Check the FIFO is empty 
   assert( data.size() == 0 );

   // Free up the resources 
   delete condition;
   condition = 0;
   delete mutex;
   mutex = 0;
}


template<class Msg> void
Fifo<Msg>::add( const Msg& msg )
{
   int error;
   
   // lock the object 
   error = mutex->lock();
   assert( !error );
   
   // modify the data structure 
   data.push_back( msg );

   // tell any waiting threads the FIFO is no longer empty 
   error = condition->signal();
   assert( !error );

   // unlock the object 
   error = mutex->unlock();
   assert( !error );
}


template<class Msg> FifoEventId
Fifo<Msg>::addDelayMs(const Msg& msg,const int delayMs)
{
   struct FifoEvent baseMsg;

   // copy msg on heap 
   Msg* newMsg = new Msg;
   *newMsg = msg;

   // set up the baseMsg
   baseMsg.msg = newMsg;

   // add the event to the base classes event list 
   FifoEventId id = addEvent( baseMsg , delayMs );

   return id;
}


template<class Msg> int
Fifo<Msg>::size() const
{
   int error;
   
   // lock the object - needed data structure could change during data.size()
   error = mutex->lock();
   assert( !error );

   // get the size 
   int returnValue = data.size();

   // unlock the object 
   error = mutex->unlock();
   assert( !error );

   return returnValue;
}


template<class Msg> Msg
Fifo<Msg>::getNext()
{
   int error;
   
   // lock the object 
   error = mutex->lock();
   assert( !error );
   
   // wait for the FIFO to have something in it 
   while ( data.size() == 0 )
   {
      // this will unlock the object, wait for a signal, and relock
       error = condition->wait( mutex );
       assert( !error );
   }
   assert( data.size() > 0 );
   
   // get the item off the list 
   Msg returnVal = data.front();
   data.pop_front();

   // unlock the object 
   error = mutex->unlock();
   assert( !error );

   return returnVal;
}


template<class Msg>  void 
Fifo<Msg>::add(const FifoEvent& baseMsg)
{
   assert( this == baseMsg.fifo );
   assert( baseMsg.msg );

   // cast the message back to the correct type 
   Msg* message = static_cast<Msg*>( baseMsg.msg );
   assert( message );

   // put the message on the Fifo 
   add( *message );
   
   // free the memory for the message 
   delete message;  message=NULL;
}

    
template<class Msg>  void 
Fifo<Msg>::deleteMsg(FifoEvent& baseMsg)
{
   assert( this == baseMsg.fifo );
   assert( baseMsg.msg );

   // cast the message back to the correct type 
   Msg* message = static_cast<Msg*>( baseMsg.msg );
   assert( message );

   // free the memory for the message 
   delete message;  message=NULL; baseMsg.msg=NULL;
}

#endif // fifo.h

