/*********************************************************************

 $Id: Sptr.hxx,v 1.17 2000/06/07 02:50:24 bogawa Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************

 $Log: Sptr.hxx,v $
 Revision 1.17  2000/06/07 02:50:24  bogawa
 enhanced Smart Pointer code

 Revision 1.16  2000/04/03 22:59:02  bogawa
 fix for smart pointer possible race

 Revision 1.15  2000/02/24 18:32:22  bogawa
 fixes for VxWorks Compile

 Revision 1.14  2000/02/15 19:51:23  bogawa

 update with the new  VTest class

 Revision 1.13  2000/02/09 22:50:51  bogawa
 smart pointer debugging support fixes

 Revision 1.12  2000/01/12 23:48:06  bogawa
 fixes for minor issues in the Sptr and Fifo code

 Revision 1.11  2000/01/11 00:56:00  bogawa
 copyright notice fixes for year 2000

 Revision 1.10  1999/12/21 21:51:46  bogawa
 fixes for the great renaming

 Revision 1.9  1999/11/30 02:09:31  eckelcu
 Modified Files:
 	Data.cxx Data.hxx DataBuff.cxx DataBuff.hxx cpLog.cxx
 	debug.cxx fifo.cxx parse3tuple.cxx parse3tupleTest.cxx
 	sptr.hxx support.cxx support.hxx timer.cxx udp.cxx udp.hxx
 	udpTransport.cxx udpTransport.hxx vcondition-linux.cxx
 	vcondition-vxworks.cxx vmutex-linux.cxx vmutex-vxworks.cxx
 	vthread-linux.cxx vthread-vxworks.cxx vtime-vxworks.cxx
 Added consistent cvs header and LGPL license info to all files in /util

    
 ********************************************************************/

#ifndef SPTR_HXX_
#define SPTR_HXX_

#define SPTR_DEBUG 1
#define SPTR_TRACEBACK 1

#include "VMutex.h"
#include "debug.h"

#if SPTR_DEBUG

#include <map>

/// this is a debugging class for pointers

class SptrDebugger
{
private:
    SptrDebugger();

    static SptrDebugger* getInstance();

    typedef map <void*, bool> PtrMap;

    static SptrDebugger* impl_;

    PtrMap ptrsInUse;

    VMutex mutex;

public:
    static void newPtr(void* ptr);

    static void deletePtr(void* ptr);
};


void sptrDebugMarkTraceback(void* sptr, void* ptr);
void sptrDebugClearTraceback(void* sptr, void* ptr);
void sptrDebugDumpTraceback(char* filename);

#endif


/** Template simulates a pointer with a "smart" reference that deletes
    the item when no more references exist
*/
template<class T> 
class Sptr
{
private:

    T* ptr;
    int* count;
    VMutex* mutex;

//    pthread_mutex_t* mutex;

    void increment() {
	if(ptr)
	{
	    if(!count) 
	    {
		count = new int(0);
	    }
	    if(!mutex)
	    {
		mutex = new VMutex;
	    }

	    mutex->lock();
	    (*count)++;
	    mutex->unlock();
	}
#if SPTR_TRACEBACK
	sptrDebugMarkTraceback((void*) this, (void*) ptr);
#endif

    }

    void decrement() {
	if(ptr && count)
	{
#if SPTR_TRACEBACK
	    sptrDebugClearTraceback((void*) this, (void*) ptr);
#endif
            bool countIsZero;
	    mutex->lock();
	    (*count)--;
	    countIsZero = (*count == 0);
	    mutex->unlock();
	    if(countIsZero)
	    {
#if SPTR_DEBUG
#if 0
		cerr << "deleting " << ptr << endl;
		breakpoint();
#endif
		SptrDebugger::deletePtr((void*) ptr);
#endif
		delete ptr; ptr = 0;
		delete count; count = 0;
		delete mutex; mutex = 0;
	    }
	}
	ptr = 0;
	count = 0;
	mutex = 0;
    }

public:
    template<class T2>
    operator Sptr<T2> () { return Sptr<T2>(ptr, count, mutex); }

    Sptr() : ptr(0), count(0), mutex(0) {};

    Sptr(T* original, int* myCount=0, VMutex* myMutex=0)
	: ptr(original), count(myCount), mutex(myMutex)
	{
	    if(ptr)
	    {
#if SPTR_DEBUG
		if(!myCount || !myMutex)
		{
		    SptrDebugger::newPtr((void*) ptr); // void* takes away warning
		}
#endif
		increment();
	    }
	};

    Sptr(const Sptr& x)
	: ptr(x.ptr), count(x.count), mutex(x.mutex)
	{
	    increment();
	};

    ~Sptr() 
	{
	    {
		decrement();
	    }
	}

    T& operator*() const {
	assert(ptr);
	return *ptr;
    }

    int operator!() const {
	if(ptr)
	{
	    return (ptr == 0);
	}
	else
	    return true;
    }

    T* operator->() const {
	return ptr;
    }

    template<class T2> Sptr& dynamicCast(const Sptr<T2>& x) 
	{
	    if (ptr == x.getPtr()) return *this;
	    decrement();
	    if(T* p = dynamic_cast<T*>(x.getPtr())) {
		count = x.getCount();
		mutex = x.getMutex();
		ptr = p;
		increment();
	    }
	    return *this;
	}

     template<class T2>
     Sptr& operator=(const Sptr<T2>& x) 
 	{
 	    if (ptr == x.getPtr()) return *this;
	    decrement();
	    ptr = x.getPtr();
	    count = x.getCount();
	    mutex = x.getMutex();
	    increment();
 	    return *this;
 	}


     Sptr&    operator=(T* original)
	{
	    if (ptr == original) return *this;
	    decrement();
#if SPTR_DEBUG
	    SptrDebugger::newPtr((void*) original);
#endif
	    ptr = original;
	    increment();
	    return *this;
	};
     Sptr&    operator=(const Sptr& x)
	{
	    if (ptr == x.getPtr()) return *this;
	    decrement();
	    ptr = x.ptr;
	    count = x.count;
	    mutex = x.mutex;
	    increment();
	    return *this;
	}



    friend int operator!=(const void* y, const Sptr& x)
	{
	    if(x.ptr != y)
		return true;
	    else
		return false;
	}

    friend int operator!=(const Sptr& x, const void* y)
	{
	    if(x.ptr != y)
		return true;
	    else
		return false;
	}

    friend int operator==(const void* y, const Sptr& x)
	{
	    if(x.ptr == y)
		return true;
	    else
		return false;
	}

    friend int operator==(const Sptr& x, const void* y)
	{
	    if(x.ptr == y)
		return true;
	    else
		return false;
	}

    int operator==(const Sptr& x)
	{
	    if(x.ptr == ptr)
		return true;
	    else
		return false;
	}

    // the following interfaces are here because they may sometimes be
    // necessary.  DO NOT USE unless you are sure of what you are doing.

    VMutex* getMutex() const {
	return mutex;
    }

    int* getCount() const {
	return count;
    }

    T* getPtr() const {
	return ptr;
    }

};


#endif
