/*********************************************************************

 $Id: Timer.h,v 1.14 2000/04/16 23:49:04 hong Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************

 $Log: Timer.h,v $
 Revision 1.14  2000/04/16 23:49:04  hong
 Added support for provisional and final responses

 Revision 1.13  2000/01/11 00:56:00  bogawa
 copyright notice fixes for year 2000

 Revision 1.12  1999/12/24 02:24:45  tina
 Added consistent cvs header and LGPL license info.

    
 ********************************************************************/

#ifndef TIMER_H_
#define TIMER_H_

#include "debug.h"
#include "VMutex.h"
#include "VTime.hxx"
#include <unistd.h>

#include "global.h"
#include <list>

bool operator < ( const struct timeval& lhs, const struct timeval& rhs);
void addMs ( struct timeval* lhs, int timeMs);
void subMs ( struct timeval* lhs, int timeMs);

long tv_diff ( const struct timeval& lhs, const struct timeval& rhs);


typedef unsigned int TimerEventId ;

template<class X>
class TimerObj
{
public:
    TimerObj() {}

    TimerObj(const struct TimerObj& x) :
	myTime(x.myTime),
	myObject(x.myObject),
	eventId(x.eventId) {}

    TimerObj(const struct timeval& time, X object) : 
	myTime(time), 
	myObject(object),
	eventId(0)
	{}

    // assignment operator
    void operator = (const struct TimerObj& x) {
      myTime=x.myTime; myObject=x.myObject; eventId=x.eventId;
    }

    // equality check operator
    bool operator == (const struct TimerObj& x) {
      if ((myTime.tv_sec==x.myTime.tv_sec) && (myTime.tv_usec==x.myTime.tv_usec) && (myObject==x.myObject) && (eventId==x.eventId))
          return( true );
      return( false );
    }

    struct timeval myTime;
    X myObject;
    TimerEventId eventId;
};

template<class X> bool operator < (const TimerObj<X>& x, const TimerObj<X>& y)
{
    return x.myTime < y.myTime; 
}

template<class T>
class Timer
{
public:
    Timer();
    Timer(long usecWait, T bad);
    Timer(const struct timeval& minWait, T bad);
    void insert(T obj, int msDelay);
    void remove(const struct TimerObj<T>& obj);
    TimerObj<T> getlastInserted() { return lastInserted; }
    bool sleepFor(struct timeval* t);

    bool ready();
    T top();
    void pop();
private:
    TimerEventId nextEventId;
    list<TimerObj<T> > events;
//    priority_queue<TimerObj<T> > events;
    VMutex* timerMutex;
    struct timeval myMinWait;
    T bogus;

    static TimerObj<T> lastInserted;
};

template<class T> TimerObj<T> Timer<T>::lastInserted = TimerObj<T>();

template<class T>
Timer<T>::Timer()
    : nextEventId(0)
{
    timerMutex = new VMutex;

    myMinWait.tv_sec = 0;
    myMinWait.tv_usec = 0;

    TimerObj<T> foo;

    events.push_back(foo);
    pop();
}

template<class T>
Timer<T>::Timer(long usecWait, T bad)
    : nextEventId(0),
      bogus(bad)
{
    timerMutex = new VMutex;

    myMinWait.tv_sec = 0;
    myMinWait.tv_usec = usecWait;
    
    while(myMinWait.tv_usec > 1000000)
    {
	myMinWait.tv_sec++;
	myMinWait.tv_usec -= 1000000;
    }

    TimerObj<T> foo;

    events.push_back(foo);
    pop();
}

template<class T>
Timer<T>::Timer(const struct timeval& minWait, T bad)
    : nextEventId(0),
      myMinWait(minWait),
      bogus(bad)
{
    timerMutex = new VMutex;

    TimerObj<T> foo;

    events.push_back(foo);
    pop();
}



template<class T>
bool Timer<T>::sleepFor(struct timeval* t)
{
    struct timeval now;
    bool retval;

    int err = gettimeofday(&now, NULL);
    assert( !err );

    int error = timerMutex->lock();
    if(!events.empty())
    {
	if(events.front().myTime < now)
	{
	    // indicate almost no time to wait!
	    t->tv_sec = 0;
	    t->tv_usec = 1;
	    XDEBUG(3, cout << "diff: " << tv_diff(events.front().myTime, now) << endl;) ;
	    retval = false;
	}
	else
	{
	    t->tv_sec = events.front().myTime.tv_sec - now.tv_sec;
	    t->tv_usec = events.front().myTime.tv_usec - now.tv_usec;
	    while(t->tv_usec < 0)
	    {
		t->tv_sec--;
		t->tv_usec += 1000000;
	    }
	    retval = true;
	}
    } 
    else
    {
	*t = myMinWait;
	retval = true;
    }
    error = timerMutex->unlock();
    return retval;
}


template<class T>
void Timer<T>::insert(T obj, int msDelay)
{
   assert( INT_MAX >  2000000000 );
   assert( INT_MIN < -2000000000 );

   struct timeval now;
   int err = gettimeofday(&now, NULL);
   assert( !err );

   XDEBUG(3, cout << "DELAY: " << msDelay << endl;);
   addMs(&now, msDelay);

   TimerObj<T> x(now, obj);

   // add the event to the priority queue
   
   // lock the object 
   int error = timerMutex->lock();
   if(error)
   {
	cout << "error is " << error << endl;
   }
   assert( !error );

   // get a new ID 
   TimerEventId id = nextEventId++;
   x.eventId = id;

   // put the message on the FIFO 
   events.push_back( x );
   
   // save the last inserted item
   lastInserted = x;
   
   // get the list into priority order 
   events.sort();

   // unlock the object 
   error = timerMutex->unlock();
   assert( !error );
      
}

template<class T>
void Timer<T>::remove(const struct TimerObj<T>& obj )
{
    if(!events.empty())
    {
	int error = timerMutex->lock();
	assert( !error );
;
	events.remove( obj );

	error = timerMutex->unlock();
	assert( !error );
    }
}

template<class T>
void Timer<T>::pop()
{
    if(!events.empty())
    {
	int error = timerMutex->lock();
	assert( !error );

	events.pop_front();

	error = timerMutex->unlock();
	assert( !error );
    }
}

template<class T>
bool Timer<T>::ready()
{
    struct timeval now;
    int err = gettimeofday(&now, NULL);
    assert( !err );

    if(!events.empty())
    {
	if(events.front().myTime < now)
	{
	    return true;
	}
    }

    return false;
}

template<class T>
T Timer<T>::top()
{
    if(!events.empty())
    {
	return events.front().myObject;
    }
    return bogus;
}

#endif
