/*********************************************************************

 $Id: VThread.h,v 1.14 2000/05/24 18:56:10 bogawa Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************

 $Log: VThread.h,v $
 Revision 1.14  2000/05/24 18:56:10  bogawa
 per-thread debugging for cpLog

 Revision 1.13  2000/04/20 01:22:34  bogawa
 added another thread test

 Revision 1.12  2000/01/11 00:56:00  bogawa
 copyright notice fixes for year 2000

 Revision 1.11  1999/12/30 23:58:06  eckelcu
 Modified Files:
 	VThread.h vthread-linux.cxx vthread-vxworks.cxx
 	test/VThreadTest.cxx
 Removed getHandle()  and fixed bug in exit() for linux.

 Revision 1.10  1999/12/28 04:30:30  ctam
 Change default stack size to 100k instead of 10K.  10K was too small
 and caused corruption of the stack.

 Modified Files:
     VThread.h

 Revision 1.9  1999/12/24 02:23:36  tina
 Added consistent cvs header and LGPL license info.

    
 ********************************************************************/

#ifndef VTHREAD_H_
#define VTHREAD_H_

#include "VTime.hxx"
#include "VThreadException.hxx"
    
/********************************************************************
  OS independent definitions
********************************************************************/
    
// VThread priorities
#define VTHREAD_PRIORITY_DEFAULT 0
    
// VThread stack
//#define VTHREAD_STACK_SIZE_DEFAULT 10000 // in bytes
#define VTHREAD_STACK_SIZE_DEFAULT 100000 // in bytes
    
// VThread flags
const unsigned long VTHREAD_SCHED_MASK    = 0x0000000f;
const unsigned long VTHREAD_SCHED_DEFAULT = 0x00000001;
const unsigned long VTHREAD_SCHED_FIFO    = 0x00000002;
const unsigned long VTHREAD_SCHED_RR      = 0x00000004;

/********************************************************************
  OS dependent definitions
********************************************************************/

#ifdef __vxworks   

#include <taskLib.h>
#include <sysLib.h>

// VThread types for VxWorks
typedef int vthread_t;
typedef SEM_ID vmutex_t;
typedef SEM_ID vcondition_t;
typedef unsigned int vthread_attr_t;
//VThread constants for VxWorks
const unsigned long  VTHREAD_PRIORITY_VXWORKS_DEFAULT = 150;
    
/**
 ** Suspend execution of the calling task for the specified number 
 ** of seconds
 */
int 
sleep( int seconds );

/**
 ** Suspend execution of the calling task for the specified number 
 ** of micro seconds
 */
int 
usleep( int useconds );

#else

#include <pthread.h>
// VThread types for Linux
typedef pthread_t  vthread_t;
typedef pthread_mutex_t vmutex_t;
typedef pthread_attr_t vthread_attr_t;    
typedef pthread_cond_t  vcondition_t;
// VThread constants for Linux
    
#endif // __vxworks

    
// VThread constants
const vthread_t VTHREAD_ID_INVALID = 0;    
    
/** Class VThread
 ** Defines a simple thread class which provides
 ** means of controlling a separate thread of control in which
 ** the scheduling mechanism and priority may be specified. Other
 ** potentially operating system dependent capabilites are
 ** provided in generic fashion as deemed necessary and useful.
 **
 ** Each instance of the class supports one thread of execution.
 ** Creating multiple threads requires creating multiple VThread objects.
 **
 ** The following exceptions may be thrown by methods of this class:
 ** VExceptionMemory(),
 ** VThreadExceptionInvalidAttributes(),
 ** VThreadExceptionInvalidPriority()
 **/
class VThread
{
public:
    
    /**
     ** Create an empty VThread object
     ** Note, this does NOT start a new thread of control
     ** associated with the VThread.
     */
    VThread( void );
    
    /**
     ** Delete a VThread object
     ** Note, this does not stop the thread of control
     ** associated with the VThread.
     */
    ~VThread( void );
    
    /**
     ** Creates a new thread of control, with the specified attributes,
     ** running startFunc with startArgs.
     **
     ** The flags are a bitwise-OR of the following:
     ** VTHREAD_SCHED_FIFO, VTHREAD_SCHED_RR, VTHREAD_SCHED_DEFAULT,
     **( rest not supported yet)
     ** VTHREAD_CANCEL_DISABLE, VTHREAD_CANCEL_ENABLE, 
     ** VTHREAD_CANCEL_DEFERRED, VTHREAD_CANCEL_ASYNCHRONOUS,
     ** VTHREAD_BOUND, VTHREAD_NEW_LWP, VTHREAD_DETACHED,
     ** VTHREAD_SUSPENDED, VTHREAD_DAEMON, VTHREAD_JOINABLE,
     **
     ** By default, or if priority is set to
     ** VTHREAD_PRIORITY_DEFAULT, an "appropriate" priority
     ** value for the given scheduling policy (specified in flags,
     ** e.g., VTHREAD_SCHED_DEFAULT) is used. This value is calculated
     ** dynamically, and is the median value between the minimum
     ** and maximum priority values for the given policy. If an
     ** explicit value is given, it is used. Note that actual
     ** priority values are EXTREMEMLY implementation-dependent,
     ** and are probably best avoided.
     ** 
     ** The following exceptions may be thrown by this method:
     ** VExceptionMemory(),
     ** VThreadExceptionInvalidAttributes(),
     ** VThreadExceptionInvalidPriority()
     **
     ** Returns 0, if successful, or a negative error code.
     */
    int
    spawn( void *(*startFunc)(void *),
	   void *startArgs = 0,
	   unsigned long flags = 0,
	   unsigned long priority = VTHREAD_PRIORITY_DEFAULT,
	   int stack_size = VTHREAD_STACK_SIZE_DEFAULT
	);

    /**
     ** Wait for thread to exit. Returns 0 if successful, or a
     ** negative error code.
     */
    int
    join( void **status = 0 );
    

    /** 
     ** Returns the priority of the thread
     */
    int
    getPriority( void );
    

    /**
     ** Set the priority of the thread. Returns 0 if successful, or a
     ** negative error code.
     **
     ** The following exceptions may be thrown by this method:
     ** VThreadExceptionInvalidPriority()
     **/
    int
    setPriority( int priority );


    /**
     ** Return the unique Id of the VThread
     */
    const vthread_t*
    getId( void );
    

    /**
     ** Return ptr to the os specific attributes of the VThread
     */
    const vthread_attr_t*
    getAttributes( void );

    /**
     ** Exit the current thread of control associated with the VThread 
     */
    void
    exit( void  );    

    /**
       Return the VThread ID of the current process
    */

    static const vthread_t
    selfId ( void );
    
private:
    vthread_t myId;
    vthread_attr_t myAttributes;
};

#endif // VTHREAD_H_
