#ifndef __vxworks

/*********************************************************************

 $Id: VThreadPool.cxx,v 1.3 2000/05/24 18:56:10 bogawa Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************
 $Log: VThreadPool.cxx,v $
 Revision 1.3  2000/05/24 18:56:10  bogawa
 per-thread debugging for cpLog

 Revision 1.2  2000/05/24 17:48:54  surendra
 Moved destructor in the cxx from hxx.

 Revision 1.1  2000/05/10 23:05:34  surendra
 Moved from policy/common

 Revision 1.1  2000/05/04 02:01:55  surendra

 Added threadpool object


 *********************************************************************/
#include <errno.h>
#include <stdio.h>

#include "VThreadPool.hxx"
#include "cpLog.h"

VThreadPool::VThreadPool(int numThreads)
{
   _queueClosed = 0; 
   _shutdown = 0; 
   _numThreads = numThreads;
   _currQueueSize = 0;
   init();
}

void
VThreadPool::init() throw (VThreadInitException&)
{
   int ret;
   char buf[256];
   if((ret = pthread_mutex_init(&_queueLock, NULL)) != 0)
   {
       sprintf(buf, "pthread_mutext_init %s", strerror(errno));
       cpLog(LOG_ALERT, buf);
       throw VThreadInitException(buf, __FILE__, __LINE__, errno);
   }

   if((ret = pthread_cond_init(&_queueNotEmpty, NULL)) != 0)
   {
       sprintf(buf, "pthread_cond_init %s", strerror(errno));
       cpLog(LOG_ALERT, buf);
       throw VThreadInitException(buf, __FILE__, __LINE__, errno);
   }

   if((ret = pthread_cond_init(&_queueNotFull, NULL)) != 0)
   {
       sprintf(buf, "pthread_cond_init %s", strerror(errno));
       cpLog(LOG_ALERT, buf);
       throw VThreadInitException(buf, __FILE__, __LINE__, errno);
   }

   if((ret = pthread_cond_init(&_queueEmpty, NULL)) != 0)
   {
       sprintf(buf, "pthread_cond_init %s", strerror(errno));
       cpLog(LOG_ALERT, buf);
       throw VThreadInitException(buf, __FILE__, __LINE__, errno);
   }

   ///create threads
   for(int i =0 ; i < _numThreads; i++)
   {
       VThread* thread = new VThread();
       _workerThreads.push_back(thread);
       ret = thread->spawn(workerProc, this);
       if(ret != 0)
       {
           sprintf(buf, "pthread_create %s", strerror(errno));
           cpLog(LOG_ALERT, buf);
           throw VThreadInitException(buf, __FILE__, __LINE__, errno);
       }
   }
}

void 
VThreadPool::addFunctor(const VFunctor& functor) throw (VThreadInitException&)
{
    pthread_mutex_lock(&_queueLock);
    if(_shutdown || _queueClosed)
    {
        pthread_mutex_unlock(&_queueLock);
        cpLog(LOG_DEBUG, "Pool shutting down..");
        throw VThreadInitException("Pool shutting down..",
                                __FILE__, __LINE__, 0);
    }

    if(_currQueueSize == 0)
    {
        _queue.push_back(functor);
        pthread_cond_signal(&_queueNotEmpty);
    }
    else
    {
        _queue.push_back(functor);
        pthread_cond_signal(&_queueNotEmpty);
    }
    _currQueueSize++;
    pthread_mutex_unlock(&_queueLock);
}

void 
VThreadPool::shutdown(bool finish)
{
    int ret;

    if((ret = pthread_mutex_lock(&_queueLock)) != 0)
    {
        cpLog(LOG_DEBUG, "Failed to get the lock, still shutting down..");
        exit(-1);
    }
    if(_queueClosed || _shutdown)
    {
        pthread_mutex_unlock(&_queueLock);
        return;
    }
    _queueClosed = 1;
    if(finish)
    {
        //Wait till the entire queue is finished
        while(_currQueueSize != 0)
        {
            pthread_cond_wait(&_queueEmpty, &_queueLock);
        }
    }
    _shutdown = 1;

    pthread_mutex_unlock(&_queueLock);

    //Wakeup any worker thread so that they can recheck shutdown flag
    pthread_cond_broadcast(&_queueNotEmpty);
    pthread_cond_broadcast(&_queueNotFull);
    //wait for workers to exit
    for(ListOfThreads::iterator itr = _workerThreads.begin();
             itr != _workerThreads.end(); itr++)
    {
        Sptr<VThread> vth = (*itr);
        if((ret = vth->join()) != 0)
        {
            cpLog(LOG_ALERT, "Failed to join thread (%d)", vth->getId());
        }
    }
}

void*
VThreadPool::workerProc(void* args)
{
    VFunctor* work = NULL;
    VThreadPool& self = *(static_cast<VThreadPool*>(args));
    cpLog(LOG_DEBUG, "Creating worker thread (%d)", pthread_self());
    for(;;)
    {
        pthread_mutex_lock(&(self._queueLock));
        while((self._currQueueSize == 0) && !(self._shutdown))
        { 
            pthread_cond_wait(&(self._queueNotEmpty), &(self._queueLock));
        }

        if(self._shutdown)
        {
            pthread_mutex_unlock(&(self._queueLock));
            cpLog(LOG_DEBUG, "Thread (%d) exiting.." , pthread_self());
            pthread_exit(NULL);
        }
        work = &(self._queue.front());
        self._queue.pop_front();
        self._currQueueSize--;

        if(self._currQueueSize == 0)
        { 
            pthread_cond_signal(&(self._queueEmpty));
        }
        
        pthread_mutex_unlock(&(self._queueLock));
        try
        {
            if(work)
            {
                cpLog(LOG_DEBUG, "Thread (%d) Doing work." , pthread_self());
                work->doWork(); 
            }
        }
        catch(VException& e)
        {
            cpLog(LOG_ALERT, "Work threw an exception: %s", e.getDescription().c_str());
        }
    }
}

VThreadPool::~VThreadPool()
{
    cpLog(LOG_DEBUG, "VThreadPool::~VThreadPool");
    shutdown();
}

#endif
