/*********************************************************************

 $Id: vthread-linux.cxx,v 1.14 2000/05/24 18:56:10 bogawa Exp $

 *********************************************************************

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.  In addition to the
  terms and conditions set forth in the GNU Lesser General Public
  License, as a condition of using this library you are required to
  grant to all users of this library or any implementation utilizing
  or derived from this library a reciprocal, no cost, worldwide,
  perpetual, non-exclusive, non-transferable, unrestricted license to
  your claims of all patents and patent applications throughout the
  world that are infringed by the library or any implementation
  utilizing or derived from this library.  In the event you
  redistribute this library or any implementation utilizing or derived
  from this library, you must prominently display the foregoing terms
  and conditions with the library or the implementation utilizing or
  derived from this library.

  In the event of a conflict of terms between the foregoing license
  grant and the terms set forth in the GNU Lesser General Public
  License, the foregoing terms and conditions shall be deemed to
  govern.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not; write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  Copyright 1999-2000 Vovida Networks, Inc.  All Rights Reserved.

 *********************************************************************

 $Log: vthread-linux.cxx,v $
 Revision 1.14  2000/05/24 18:56:10  bogawa
 per-thread debugging for cpLog

 Revision 1.13  2000/05/03 17:56:05  eckelcu
 Modified to work for Solaris.

 Revision 1.12  2000/05/01 19:06:14  bogawa
 fixes for FreeBSD compiling in util directory

 Revision 1.11  2000/04/20 23:30:17  icahoon
 cleaned up compile warningon selfId

 Revision 1.10  2000/04/20 01:22:34  bogawa
 added another thread test

 Revision 1.9  2000/04/10 18:24:37  eckelcu

 Modified semantics of VThread::spawn() for linux such that the default
 for scheduling policy and priority is to inherit them from the parent
 thread. This is more in line with the default for pthreads.

 Revision 1.8  2000/01/11 00:56:01  bogawa
 copyright notice fixes for year 2000

 Revision 1.7  1999/12/30 23:58:06  eckelcu
 Modified Files:
 	VThread.h vthread-linux.cxx vthread-vxworks.cxx
 	test/VThreadTest.cxx
 Removed getHandle()  and fixed bug in exit() for linux.

 Revision 1.6  1999/12/23 21:02:37  eckelcu
 	Makefile VException.cxx VException.hxx VThread.h
 	VThreadException.cxx vthread-linux.cxx vthread-vxworks.cxx
 	test/VThreadTest.cxx
 Added Files:
 	VConditionException.cxx VConditionException.hxx
 Additional error and exception handling in VThread class

 Revision 1.5  1999/12/21 21:51:46  bogawa
 fixes for the great renaming

 Revision 1.4  1999/12/14 00:31:38  eckelcu
 Modified Files:
 	Makefile vthread-linux.cxx vthread.h
 Added Files:
 	VException.cxx VException.hxx VThreadException.cxx
 	VThreadException.hxx test/Makefile test/VThreadTest.cxx
 Created base exception class, VException, and modified VThread class
 to make use of it.
 Also created a new test program which exercises the new VThread class.

 Revision 1.3  1999/12/06 22:47:36  eckelcu
 Modified Files:
 	vthread-linux.cxx vthread-vxworks.cxx vthread.h
 Removed kill() method. Should use exit() instead.

 Revision 1.2  1999/11/30 02:09:31  eckelcu
 Modified Files:
 	Data.cxx Data.hxx DataBuff.cxx DataBuff.hxx cpLog.cxx
 	debug.cxx fifo.cxx parse3tuple.cxx parse3tupleTest.cxx
 	sptr.hxx support.cxx support.hxx timer.cxx udp.cxx udp.hxx
 	udpTransport.cxx udpTransport.hxx vcondition-linux.cxx
 	vcondition-vxworks.cxx vmutex-linux.cxx vmutex-vxworks.cxx
 	vthread-linux.cxx vthread-vxworks.cxx vtime-vxworks.cxx
 Added consistent cvs header and LGPL license info to all files in /util

    
 ********************************************************************/

#ifndef __vxworks
    
#include <cassert>
#include "VThread.h"
#include "cpLog.h"

/*********************************************************************

VThread for Linux

*********************************************************************/

VThread::VThread( void )
{
    myId = VTHREAD_ID_INVALID;
}
    
VThread::~VThread( void )
{
    if( myId != VTHREAD_ID_INVALID )
    {
	// warn user but don't kill thread
	cpLog( LOG_DEBUG, "active VThread:%d going out of scope", myId );
    }
    myId = VTHREAD_ID_INVALID;
}
    
int
VThread::spawn( void *( *startFunc )( void * ),
		void *startArgs = 0,
		unsigned long flags = 0,
		unsigned long priority = VTHREAD_PRIORITY_DEFAULT,
		int stack_size = VTHREAD_STACK_SIZE_DEFAULT
    )
{
    if( myId != VTHREAD_ID_INVALID )
    {
	// return without spawning another thread
	cpLog( LOG_ERR, "VThread:%d is already active", myId );
	return -1; // type of vthread_t is not int necessarily
    }

    struct sched_param priorityParams;
    int retVal = 0;
    
    assert( pthread_attr_init(&myAttributes) == 0 );
    
    // extract scheduling preference from flags and set scheduling
    // policy accordingly
    switch( flags & VTHREAD_SCHED_MASK )
    {
	case VTHREAD_SCHED_FIFO:
	{    
	    assert( pthread_attr_setinheritsched( &myAttributes,
						  PTHREAD_EXPLICIT_SCHED )
		    == 0 );
	    retVal = pthread_attr_setschedpolicy( &myAttributes,
						  SCHED_FIFO );
	    if( retVal )
	    {
		throw VThreadExceptionInvalidAttributes(
		    "cannot set schedule as VTHREAD_SCHED_FIFO",
		    __FILE__,
		    __LINE__,
		    retVal );
	    }
	}
	break;
    
	case VTHREAD_SCHED_RR:
	{
	    assert( pthread_attr_setinheritsched( &myAttributes,
						  PTHREAD_EXPLICIT_SCHED )
		    == 0 );
	    retVal = pthread_attr_setschedpolicy( &myAttributes,
						  SCHED_RR );
	    if( retVal )
	    {
		throw VThreadExceptionInvalidAttributes(
		    "cannot set schedule as VTHREAD_SCHED_RR",
		    __FILE__,
		    __LINE__,
		    retVal );
	    }
	}
	break;

        case VTHREAD_SCHED_DEFAULT:    
	default:
	{
	    // PTHREAD_INHERIT_SCHED is not supported by Solaris
	    // and is not required for Linux so it has been removed
	    
	    // inherit scheduling policy of parent as default
	    //assert( pthread_attr_setinheritsched( &myAttributes,
	    //                                      PTHREAD_INHERIT_SCHED )
	    //    == 0 );
	}
	break;
    }
    
    // if anything expect default, set scheduling priority explicitly;
    // note that by default the priority of the parent thread is inherited
    if( priority != VTHREAD_PRIORITY_DEFAULT )
    {
	// probably should improve to use relative values
	priorityParams.sched_priority = priority;
	retVal = pthread_attr_setschedparam( &myAttributes,
					     &priorityParams );
	if( retVal )
	{
	    throw VThreadExceptionInvalidPriority(
		"cannot set priority",
		__FILE__,
		__LINE__,
		retVal );
	}
    }
    
    // spawn the thread
    return( pthread_create( &myId, &myAttributes, startFunc, startArgs ) );
}

    
int
VThread::join( void **status = 0 )
{
    int retVal = pthread_join( myId, status );
    myId = VTHREAD_ID_INVALID;
    return retVal;
}
    

int
VThread::getPriority( void )
{
    struct sched_param priorityParams;
    
    assert( pthread_attr_getschedparam( &myAttributes,
					&priorityParams ) == 0 );
    return priorityParams.sched_priority;
}


int
VThread::setPriority( int priority )
{
    struct sched_param priorityParams;
    int policy;
    int retVal;
    
    retVal = pthread_attr_setschedparam( &myAttributes, &priorityParams );
    if( retVal )
    {
	throw VThreadExceptionInvalidPriority(
	    "cannot set priority",
	    __FILE__,
	    __LINE__,
	    retVal );
    }
    
    retVal = pthread_attr_getschedpolicy( &myAttributes, &policy );
    if( retVal )
    {
	throw VThreadExceptionInvalidPriority(
	    "cannot set policy",
	    __FILE__,
	    __LINE__,
	    retVal );
    }
    return( pthread_setschedparam( myId, policy, &priorityParams ) );
}


const vthread_t*
VThread::getId( void )
{
    return &myId;
}

    
const vthread_attr_t*
VThread::getAttributes( void )
{
    return &myAttributes;
}


void
VThread::exit( void )
{
    if( myId != VTHREAD_ID_INVALID )
    {
	pthread_cancel( myId );
	myId = VTHREAD_ID_INVALID;
    }
}


const vthread_t
VThread::selfId( void )
{
    return pthread_self();
}

#endif // not __vxworks
